#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include "browse.h"

static int strcasecmp_void(const void *s1, const void*s2)
{
  return strcasecmp((const char *) s1, (const char *) s2);
}

static int bn_noext_comp(const void *s1, const void*s2)
{
  return strcmp((const char *) s1, ((const struct fileinfo*) s2)->bn_noext);
}

static int relfn_noext_comp(const void *s1, const void*s2)
{
  return strcmp((const char *) s1, ((const struct fileinfo*) s2)->relfn_noext);
}

static void free_fileinfo_void(void *p)
{
  free_fileinfo((struct fileinfo*) p);
}

static void free_noeffect(void *p)
{
  p = p;
}

#define MAXBUF 4096

static int texdeps(const struct fileinfo* fi,
		   const char* src,
		   const char* dst)
{
  FILE		 *f = fopen(fi->absfn, "r");
  char		 *s = malloc(MAXBUF * sizeof(char));
  char		 *s_abs = malloc(MAXBUF * sizeof(char));
  struct stat	 sb;
  char		 c;
  int k_ini, k = 0;
  int state = 0;
  while ((c = fgetc(f)) != EOF)
    {
      if (state < 11 && c == '\\')
	state = 1;
      else
	switch (state)
	  {
	    case 1:
	      if (c == 'i')
		state = 2;
	      else
		state = 0;
	      break;
	    case 2:
	      if (c == 'n')
		state = 3;
	      else
		state = 0;
	      break;
	    case 3:
	      if (c == 'p')
		state = 4;
	      else
		if (c == 'c')
		  state = 6;
		else
		  state = 0;
	      break;
	    case 4:
	      if (c == 'u')
		state = 5;
	      else
		state = 0;
	      break;
	    case 5:
	      if (c == 't')
		state = 10;
	      else
		state = 0;
	      break;
	    case 6:
	      if (c == 'l')
		state = 7;
	      else
		state = 0;
	      break;
	    case 7:
	      if (c == 'u')
		state = 8;
	      else
		state = 0;
	      break;
	    case 8:
	      if (c == 'd')
		state = 9;
	      else
		state = 0;
	      break;
	    case 9:
	      if (c == 'e')
		state = 10;
	      else
		state = 0;
	      break;
	    case 10:
	      if (c == '{')
		{
		  state = 11;
		  k = -1;
		}
	      else
		state = 0;
	      break;
	    case 11:
	      if (c != '}')
		{
		  if (k == -1)
		    {
		      if (c == '/')
			k = k_ini = 0;
		      else
			k = k_ini = 0;
		    }
		  s[k] = c;
		  s[k + 1] = '\0';
		  if (k + 1 < MAXBUF)
		    ++k;
		}
	      else
		{
		  if (k <= 4 || strcasecmp(s + k - 4, ".tex"))
		    if (k + 4 < MAXBUF)
		      {
			s[k] = '.';
			s[k+1] = 't';
			s[k+2] = 'e';
			s[k+3] = 'x';
			s[k+4] = '\0';
			k = k + 4;
		      }
		  s_abs = strcpy(s_abs, src);
		  s_abs = strcat(s_abs, "/");
		  s_abs = strcat(s_abs, s);
		  if (!(lstat(s_abs, &sb) != -1 && S_ISREG(sb.st_mode)))
		    {
		      fprintf(stderr, "Undefined reference to tex file '%s' in %s\n",
			      s + k_ini,
			      fi->relfn);
		    }
		  else
		    printf(" %s/%s", dst, s);
		  k = -1;
		  state = 0;
		}
	      break;
	    default:
	      ;
	  }
    }
  fclose(f);
  free(s);
  return 1;
}

static int stydeps(const struct fileinfo* fi, const char* dst)
{
  FILE		 *f = fopen(fi->absfn, "r");
  char		 *s = malloc(MAXBUF * sizeof(char));
  char		 *s_abs = malloc(MAXBUF * sizeof(char));
  struct stat	 sb;
  char		 c;
  int k_ini, k = 0;
  int state = 0;

  while ((c = fgetc(f)) != EOF)
    {
      if (state < 11 && c == '\\')
	state = 1;
      else
	switch (state)
	  {
	    case 1:
	      if (c == 'u')
		state = 2;
	      else
		state = 0;
	      break;
	    case 2:
	      if (c == 's')
		state = 3;
	      else
		state = 0;
	      break;
	    case 3:
	      if (c == 'e')
		state = 4;
	      else
		state = 0;
	      break;
	    case 4:
	      if (c == 'p')
		state = 5;
	      else
		state = 0;
	      break;
	    case 5:
	      if (c == 'a')
		state = 6;
	      else
		state = 0;
	      break;
	    case 6:
	      if (c == 'c')
		state = 7;
	      else
		state = 0;
	      break;
	    case 7:
	      if (c == 'k')
		state = 8;
	      else
		state = 0;
	      break;
	    case 8:
	      if (c == 'a')
		state = 9;
	      else
		state = 0;
	      break;
	    case 9:
	      if (c == 'g')
		state = 10;
	      else
		state = 0;
	      break;
	    case 10:
	      if (c == 'e')
		state = 11;
	      else
		state = 0;
	      break;
	    case 11:
	      if (c == '{')
		{
		  state = 12;
		  k = -1;
		}
	      else
		state = 0;
	      break;
	    case 12:
	      if (c != '}')
		{
		  if (k == -1)
		    {
		      if (c == '/')
			k = k_ini = 0;
		      else
			{
			  s = strcpy(s, fi->reldn);
			  s = strcat(s, "/");
			  k = k_ini = strlen(s);
			}
		    }
		  s[k] = c;
		  s[k + 1] = '\0';
		  if (k + 1 < MAXBUF)
		    ++k;
		}
	      else
		{
		  if (k <= 4 || strcasecmp(s + k - 4, ".sty"))
		    if (k + 4 < MAXBUF)
		      {
			s[k] = '.';
			s[k+1] = 's';
			s[k+2] = 't';
			s[k+3] = 'y';
			s[k+4] = '\0';
			k = k + 4;
		      }
		  s_abs = strcpy(s_abs, fi->dn);
		  s_abs = strcat(s_abs, "/");
		  s_abs = strcat(s_abs, s+strlen(fi->reldn)+1);
		  if ((lstat(s_abs, &sb) != -1 && S_ISREG(sb.st_mode)))
		    printf(" %s/%s", dst, s + k_ini);
		  k = -1;
		  state = 0;
		}
	      break;
	    default:
	      ;
	  }
    }
  fclose(f);
  free(s);
  return 1;
}

static int bibdeps(const struct fileinfo* fi, const char* dst)
{
  FILE		 *f = fopen(fi->absfn, "r");
  char		 *s = malloc(MAXBUF * sizeof(char));
  char		 *s_abs = malloc(MAXBUF * sizeof(char));
  struct stat	 sb;
  char		 c;
  int k_ini, k = 0;
  int state = 0;

  dst = dst;

  while ((c = fgetc(f)) != EOF)
    {
      if (state < 11 && c == '\\')
	state = 1;
      else
	switch (state)
	  {
	    case 1:
	      if (c == 'b')
		state = 2;
	      else
		state = 0;
	      break;
	    case 2:
	      if (c == 'i')
		state = 3;
	      else
		state = 0;
	      break;
	    case 3:
	      if (c == 'b')
		state = 4;
	      else
		state = 0;
	      break;
	    case 4:
	      if (c == 'l')
		state = 5;
	      else
		state = 0;
	      break;
	    case 5:
	      if (c == 'i')
		state = 6;
	      else
		state = 0;
	      break;
	    case 6:
	      if (c == 'o')
		state = 7;
	      else
		state = 0;
	      break;
	    case 7:
	      if (c == 'g')
		state = 8;
	      else
		state = 0;
	      break;
	    case 8:
	      if (c == 'r')
		state = 9;
	      else
		state = 0;
	      break;
	    case 9:
	      if (c == 'a')
		state = 10;
	      else
		state = 0;
	      break;
	    case 10:
	      if (c == 'p')
		state = 11;
	      else
		state = 0;
	      break;
	    case 11:
	      if (c == 'h')
		state = 12;
	      else
		state = 0;
	      break;
	    case 12:
	      if (c == 'y')
		state = 13;
	      else
		state = 0;
	      break;
	    case 13:
	      if (c == '{')
		{
		  state = 14;
		  k = -1;
		}
	      else
		state = 0;
	      break;
	    case 14:
	      if (c != '}')
		{
		  if (k == -1)
		    {
		      if (c == '/')
			k = k_ini = 0;
		      else
			{
			  s = strcpy(s, fi->reldn);
			  s = strcat(s, "/");
			  k = k_ini = strlen(s);
			}
		    }
		  s[k] = c;
		  s[k + 1] = '\0';
		  if (k + 1 < MAXBUF)
		    ++k;
		}
	      else
		{
		  if (k <= 4 || strcasecmp(s + k - 4, ".bib"))
		    if (k + 4 < MAXBUF)
		      {
			s[k] = '.';
			s[k+1] = 'b';
			s[k+2] = 'i';
			s[k+3] = 'b';
			s[k+4] = '\0';
			k = k + 4;
		      }
		  s_abs = strcpy(s_abs, fi->dn);
		  s_abs = strcat(s_abs, "/");
		  s_abs = strcat(s_abs, s+strlen(fi->reldn)+1);
		  if ((lstat(s_abs, &sb) != -1 && S_ISREG(sb.st_mode)))
		    printf(" %s", s+strlen(fi->reldn)+1);
		  k = -1;
		  state = 0;
		}
	      break;
	    default:
	      ;
	  }
    }
  fclose(f);
  free(s);
  free(s_abs);
  return 1;
}

static int imagesdeps(const struct fileinfo* fi, struct s_list* list_images, const char* format)
{
  FILE		 *f = fopen(fi->absfn, "r");
  char		 *s = malloc(MAXBUF * sizeof(char));
  char		 c;
  struct fileinfo* fi_img;
  int k_ini, k = 0;
  int state = 0;
  while ((c = fgetc(f)) != EOF)
    {
      if (state < 17 && c == '\\')
	state = 1;
      else
	switch (state)
	  {
	    case 1:
	      if (c == 'i')
		state = 2;
	      else
		state = 0;
	      break;
	    case 2:
	      if (c == 'n')
		state = 3;
	      else
		state = 0;
	      break;
	    case 3:
	      if (c == 'c')
		state = 4;
	      else
		state = 0;
	      break;
	    case 4:
	      if (c == 'l')
		state = 5;
	      else
		state = 0;
	      break;
	    case 5:
	      if (c == 'u')
		state = 6;
	      else
		state = 0;
	      break;
	    case 6:
	      if (c == 'd')
		state = 7;
	      else
		state = 0;
	      break;
	    case 7:
	      if (c == 'e')
		state = 8;
	      else
		state = 0;
	      break;
	    case 8:
	      if (c == 'g')
		state = 9;
	      else
		state = 0;
	      break;
	    case 9:
	      if (c == 'r')
		state = 10;
	      else
		state = 0;
	      break;
	    case 10:
	      if (c == 'a')
		state = 11;
	      else
		state = 0;
	      break;
	    case 11:
	      if (c == 'p')
		state = 12;
	      else
		state = 0;
	      break;
	    case 12:
	      if (c == 'h')
		state = 13;
	      else
		state = 0;
	      break;
	    case 13:
	      if (c == 'i')
		state = 14;
	      else
		state = 0;
	      break;
	    case 14:
	      if (c == 'c')
		state = 15;
	      else
		state = 0;
	      break;
	    case 15:
	      if (c == 's')
		state = 16;
	      else
		state = 0;
	      break;
	    case 16:
	      if (c == '{')
		{
		  state = 18;
		  k = -1;
		}
	      else
		if (c == '[')
		  state = 17;
		else
		  if (c == '*')
		    state = 16;
		  else
		  state = 0;
	      break;
	    case 17:
	      if (c == ']')
		state = 16;
	      break;
	    case 18:
	      if (c != '}')
		{
		  if (k == -1)
		    {
		      if (c == '/')
			k = k_ini = 0;
		      else
			{
			  s = strcpy(s, fi->reldn);
			  s = strcat(s, "/");
			  k = k_ini = strlen(s);
			}
		    }
		  if (c == '/')
		    k_ini = k + 1;
		  s[k] = c;
		  s[k + 1] = '\0';
		  if (k + 1 < MAXBUF)
		    ++k;
		}
	      else
		{
		  if ((fi_img = list_find_item_eq(list_images, s+1, relfn_noext_comp)))
		    printf(" $(IMAGES_DST_PATH)/%s.%s", s+1, format);
		  else
		    if ((fi_img = list_find_item_eq(list_images, s + k_ini, bn_noext_comp)))
		      printf(" $(IMAGES_DST_PATH)/%s.%s", fi_img->relfn_noext, format);
		    else
		      {
			fprintf(stderr, "Undefined reference to image '%s' in %s\n",
				s + k_ini,
				fi->relfn);
		      }
		  k = -1;
		  state = 0;
		}
	      break;
	    default:
	      ;
	  }
    }
  fclose(f);
  free(s);
  return 1;
}

static int texdocumentclass(const struct fileinfo* fi)
{
  FILE		 *f = fopen(fi->absfn, "r");
  char		 *s = malloc(MAXBUF * sizeof(char));
  char		 c;
  int state = 0;
  while ((c = fgetc(f)) != EOF)
    {
      if (state < 14 && c == '\\')
	state = 1;
      else
	switch (state)
	  {
	    case 1:
	      if (c == 'd')
		state = 2;
	      else
		state = 0;
	      break;
	    case 2:
	      if (c == 'o')
		state = 3;
	      else
		state = 0;
	      break;
	    case 3:
	      if (c == 'c')
		state = 4;
	      else
		state = 0;
	      break;
	    case 4:
	      if (c == 'u')
		state = 5;
	      else
		state = 0;
	      break;
	    case 5:
	      if (c == 'm')
		state = 6;
	      else
		state = 0;
	      break;
	    case 6:
	      if (c == 'e')
		state = 7;
	      else
		state = 0;
	      break;
	    case 7:
	      if (c == 'n')
		state = 8;
	      else
		state = 0;
	      break;
	    case 8:
	      if (c == 't')
		state = 9;
	      else
		state = 0;
	      break;
	    case 9:
	      if (c == 'c')
		state = 10;
	      else
		state = 0;
	      break;
	    case 10:
	      if (c == 'l')
		state = 11;
	      else
		state = 0;
	      break;
	    case 11:
	      if (c == 'a')
		state = 12;
	      else
		state = 0;
	      break;
	    case 12:
	      if (c == 's')
		state = 13;
	      else
		state = 0;
	      break;
	    case 13:
	      if (c == 's')
		state = 14;
	      else
		state = 0;
	      break;
	    case 14:
	      fclose(f);
	      free(s);
	      return 1;
	    default:
	      ;
	  }
    }
  fclose(f);
  free(s);
  return 0;
}

int main(int argc, char* argv[])
{
  char*		 src;
  char*		 srcimagesdir;
  char*		 dst;
  char*		 format;
  struct s_list* list;
  struct s_list* list_images;
  struct s_list* p;
  struct s_list* lext;

  if (argc != 5)
    {
      fprintf(stderr, "Usage: %s srcdir srcimagesdir dstdir format\n", argv[0]);
      exit(2);
    }
  src = argv[1];
  srcimagesdir = argv[2];
  dst = argv[3];
  format = argv[4];

  if (src[strlen(src)-1] == '/')
    src[strlen(src)-1] = '\0';
  if (dst[strlen(dst)-1] == '/')
    dst[strlen(dst)-1] = '\0';

  list = browse(argv[0], src);
  list_images = browse(argv[0], srcimagesdir);
  lext = NULL;

  for (p = list; p; p = p->next)
    {
      struct fileinfo * fi = p->data;
      list_add_item_sorted(&lext, fi->ext, strcasecmp_void);
    }

  if (lext)
    {
      printf(".SUFFIXES:");
      for (p = lext; p; p = p->next)
	if (strcasecmp((char *) p->data, ""))
	  printf(" %s", (char *) p->data);
      printf("\n\n");
    }
  if (list)
    {
      printf("target:");
      for (p = list; p; p = p->next)
	{
	  struct fileinfo * fi = p->data;
	  if (!strcasecmp(fi->ext, ".tex"))
	    if (texdocumentclass(fi) && (!strcmp(fi->dn, src) || !strcmp(fi->reldn, src)))
	      {
		if (!strcasecmp(format, "pdf"))
		  printf(" %s.pdf", fi->relfn_noext);
		if (!strcasecmp(format, "eps"))
		  printf(" %s.dvi %s.ps %s.pdf",
			 fi->relfn_noext,
			 fi->relfn_noext,
			 fi->relfn_noext);
	      }
	}
      printf("\n");
      printf("\t@echo \"Compilation success\" | $(COLORIZEGREEN);\n");
      printf("\t@printf \"\\n\";\n");
      printf("\n");
    }
  for (p = list; p; p = p->next)
    {
      struct fileinfo * fi = p->data;
      if (!strcasecmp(fi->ext, ".tex"))
	if (texdocumentclass(fi) && (!strcmp(fi->dn, src) || !strcmp(fi->reldn, src)))
	  {
	    if (!strcasecmp(format, "pdf"))
	      printf("%s.pdf: %s/%s.tex\n\n", fi->relfn_noext, dst, fi->relfn_noext);
	    if (!strcasecmp(format, "eps"))
	      printf("%s.dvi: %s/%s.tex\n\n", fi->relfn_noext, dst, fi->relfn_noext);
	  }
    }
  for (p = list; p; p = p->next)
    {
      struct fileinfo * fi = p->data;
      if (!strcasecmp(fi->ext, ".tex"))
	{
	  printf("%s/%s.tex: %s", dst, fi->relfn_noext, fi->relfn);
	  stydeps(fi, dst);
	  bibdeps(fi, dst);
	  texdeps(fi, src, dst);
	  imagesdeps(fi, list_images, format);
	  printf("\n\t@mkdir -p %s%s\n", dst, fi->reldn);
	  printf("\t@cp -f %s %s/%s.tex\n\n", fi->relfn, dst, fi->relfn_noext);
	}
      if (!strcasecmp(fi->ext, ".sty"))
	{
	  printf("%s/%s.sty: %s", dst, fi->relfn_noext, fi->relfn);
	  stydeps(fi, dst);
	  bibdeps(fi, dst);
	  texdeps(fi, src, dst);
	  imagesdeps(fi, list_images, format);
	  printf("\n\t@mkdir -p %s%s\n", dst, fi->reldn);
	  printf("\t@cp -f %s %s/%s.sty\n\n", fi->relfn, dst, fi->relfn_noext);
	}
    }
  list_free(&list, free_fileinfo_void);
  list_free(&lext, free_noeffect);

  return EXIT_SUCCESS;
}
